#ifndef STIM_CUDA_VOTE_SHARED_H
#define STIM_CUDA_VOTE_SHARED
# include <iostream>
# include <cuda.h>
#include <stim/cuda/cudatools.h>
#include <stim/cuda/sharedmem.cuh>
#include "cpyToshare.cuh"

namespace stim{
	namespace cuda{

		// this kernel calculates the vote value by adding up the gradient magnitudes of every voter that this pixel is located in their voting area
		template<typename T>
		__global__ void cuda_vote(T* gpuVote, T* gpuGrad, T* gpuTable, T phi, int rmax, int x, int y){

			//generate a pointer to shared memory (size will be specified as a kernel parameter)
			extern __shared__ float s_grad[];

			//calculate the start point for this block
			int bxi = blockIdx.x * blockDim.x;
			
			// calculate the 2D coordinates for this current thread.
			int xi = bxi + threadIdx.x;
			int yi = blockIdx.y * blockDim.y + threadIdx.y;
			// convert 2D coordinates to 1D
			int i = yi * x + xi;
						
			// define a local variable to sum the votes from the voters
			float sum = 0;
			
			//calculate the width of the shared memory block
			int swidth = 2 * rmax + blockDim.x;
			
			// compute the size of window which will be checked for finding the proper voters for this pixel
			int x_table = 2*rmax +1;
			int rmax_sq = rmax * rmax;
			int tx_rmax = threadIdx.x + rmax;
			int bxs = bxi - rmax;
			
			//for every line (along y)
			for(int yr = -rmax; yr <= rmax; yr++){
				if (yi+yr<y && yi+yr>=0){
					//copy the portion of the image necessary for this block to shared memory
					__syncthreads();
					cpyG2S1D2ch<float>(s_grad, gpuGrad, bxs, yi + yr , 2*swidth, 1, threadIdx, blockDim, x, y);
					__syncthreads();
				
					if(xi < x && yi < y){

						for(int xr = -rmax; xr <= rmax; xr++){
					
								//find the location of this voter in the atan2 table
								int id_t = (yr + rmax) * x_table + xr + rmax;

								// calculate the angle between the pixel and the current voter in x and y directions
								float atan_angle = gpuTable[id_t];
												
								// calculate the voting direction based on the grtadient direction
								int idx_share = xr + tx_rmax ;
								float theta = s_grad[idx_share*2];
								float mag = s_grad[idx_share*2 + 1];
							

								// check if the current voter is located in the voting area of this pixel.
								if (((xr * xr + yr *yr)< rmax_sq) && (abs(atan_angle - theta) <phi)){
									sum += mag;		

								}
						}
				
					}
				}
			}
			if(xi < x && yi < y)
				gpuVote[i] = sum;
			
		}

		template<typename T>
		void gpu_vote(T* gpuVote, T* gpuGrad, T* gpuTable, T phi, unsigned int rmax, unsigned int x, unsigned int y){

							
			unsigned int max_threads = stim::maxThreadsPerBlock();
			dim3 threads(max_threads, 1);
			dim3 blocks(x/threads.x + (x %threads.x == 0 ? 0:1) , y);
			
					
			// specify  share memory
			unsigned int share_bytes = (2*rmax + threads.x)*1*2*sizeof(T);
			
			//call the kernel to do the voting
			cuda_vote <<< blocks, threads,share_bytes >>>(gpuVote, gpuGrad, gpuTable, phi, rmax, x , y);

		}


		template<typename T>
		void cpu_vote(T* cpuVote, T* cpuGrad,T* cpuTable, T phi, unsigned int rmax, unsigned int x, unsigned int y){

			//calculate the number of bytes in the array
			unsigned int bytes = x * y * sizeof(T);

			//calculate the number of bytes in the atan2 table
			unsigned int bytes_table = (2*rmax+1) * (2*rmax+1) * sizeof(T);

			//allocate space on the GPU for the Vote Image
			T* gpuVote;
			cudaMalloc(&gpuVote, bytes);		

			//allocate space on the GPU for the input Gradient image
			T* gpuGrad;
			HANDLE_ERROR(cudaMalloc(&gpuGrad, bytes*2));

			//copy the Gradient Magnitude data to the GPU
			HANDLE_ERROR(cudaMemcpy(gpuGrad, cpuGrad, bytes*2, cudaMemcpyHostToDevice));

			//allocate space on the GPU for the atan2 table
			T* gpuTable;
			HANDLE_ERROR(cudaMalloc(&gpuTable, bytes_table));

			//copy the atan2 values to the GPU
			HANDLE_ERROR(cudaMemcpy(gpuTable, cpuTable, bytes_table, cudaMemcpyHostToDevice));
						
			//call the GPU version of the vote calculation function
			gpu_vote<T>(gpuVote, gpuGrad, gpuTable, phi, rmax, x , y);
							
			//copy the Vote Data back to the CPU
			cudaMemcpy(cpuVote, gpuVote, bytes, cudaMemcpyDeviceToHost) ;

			//free allocated memory
			cudaFree(gpuTable);
			cudaFree(gpuVote);
			cudaFree(gpuGrad);
		}
		
	}
}

#endif